import pytest

from aioafero import util

from .v1 import utils

thermostat = utils.create_devices_from_data("thermostat.json")[0]


@pytest.mark.parametrize(
    "vals, percentage, expected, err",
    [
        ([], None, None, True),
        ([1, 2, 3], 50, 2, False),
        ([1, 2, 3], 101, 3, False),
    ],
)
def test_percentage_to_ordered_list_item(vals, percentage, expected, err):
    if not err:
        assert util.percentage_to_ordered_list_item(vals, percentage) == expected
    else:
        with pytest.raises(ValueError):
            util.percentage_to_ordered_list_item(vals, percentage)


@pytest.mark.parametrize(
    "vals, value, expected, err",
    [
        ([1, 2, 3], 4, None, True),
        ([1, 2, 3], 2, 66, False),
    ],
)
def test_ordered_list_item_to_percentage(vals, value, expected, err):
    if not err:
        assert util.ordered_list_item_to_percentage(vals, value) == expected
    else:
        with pytest.raises(ValueError):
            util.ordered_list_item_to_percentage(vals, value)


@pytest.mark.parametrize(
    "range_vals, expected",
    [
        ({"range": {"min": 100, "max": 100, "step": 1}}, [100]),
        ({"range": {"min": 0, "max": 100, "step": 1}}, list(range(0, 101, 1))),
        ({"range": {"min": 0, "max": 100, "step": 3}}, list(range(0, 100, 3)) + [100]),
    ],
)
def test_process_range(range_vals, expected):
    assert util.process_range(range_vals) == expected


@pytest.mark.parametrize(
    "functions, func_class, func_instance, expected",
    [
        # None
        ([], "cool", None, []),
        (
            thermostat.functions,
            "temperature",
            "safety-mode-max-temp",
            [
                29.5,
                30.0,
                30.5,
                31.0,
                31.5,
                32.0,
                32.5,
                33.0,
                33.5,
                34.0,
                34.5,
                35.0,
                35.5,
                36.0,
                36.5,
                37,
            ],
        ),
        (thermostat.functions, "temperature-units", None, ["fahrenheit", "celsius"]),
        # Not processable
        (thermostat.functions, "preset", "preset-1", []),
    ],
)
def test_process_function(functions, func_class, func_instance, expected):
    assert (
        util.process_function(functions, func_class, func_instance=func_instance)
        == expected
    )


@pytest.mark.parametrize(
    "celsius, fahrenheit",
    [
        (22, 72),
        (23, 73),
        (24, 75),
        (24.5, 76),
    ],
)
def test_calculate_hubspace_fahrenheit(celsius, fahrenheit):
    assert util.calculate_hubspace_fahrenheit(celsius) == fahrenheit


@pytest.mark.parametrize(
    "celsius, fahrenheit",
    [
        (22, 72),
        (23, 73),
        (24, 75),
        (24.5, 76),
    ],
)
def test_calculate_hubspace_celsius(celsius, fahrenheit):
    assert util.calculate_hubspace_celsius(fahrenheit) == celsius



"""Base controller for climate devices."""

from typing import TypeVar

from aioafero import device
from aioafero.device import AferoDevice, AferoResource, AferoState
from aioafero.errors import DeviceNotFound
from aioafero.util import calculate_hubspace_celsius, process_function
from aioafero.v1.models import features

from .base import BaseResourcesController

AferoResourceT = TypeVar("AferoResourceT", bound=AferoResource)


def generate_target_temp(
    func_def: dict, state: AferoState
) -> features.TargetTemperatureFeature:
    """Determine the target temp based on the function definition."""
    return features.TargetTemperatureFeature(
        value=round(state.value, 2),
        step=func_def["range"]["step"],
        min=func_def["range"]["min"],
        max=func_def["range"]["max"],
        instance=state.functionInstance,
    )


class ClimateController(BaseResourcesController[AferoResourceT]):
    """Base controller for climate devices."""

    async def initialize_climate_elem(self, afero_device: AferoDevice) -> dict:
        """Initialize the climate elements of a device."""
        climate_data = {
            "available": False,
            "display_celsius": True,
            "current_temperature": None,
            "hvac_mode": None,
            "target_temperature_cooling": None,
            "target_temperature_heating": None,
            "target_temperature_auto_cooling": None,
            "target_temperature_auto_heating": None,
            "safety_max_temp": None,
            "safety_min_temp": None,
            "numbers": {},
            "selects": {},
            "sensors": {},
            "binary_sensors": {},
        }

        for state in afero_device.states:
            func_def = device.get_function_from_device(
                afero_device.functions, state.functionClass, state.functionInstance
            )
            if state.functionClass == "temperature":
                if state.functionInstance == "current-temp":
                    climate_data["current_temperature"] = (
                        features.CurrentTemperatureFeature(
                            temperature=round(state.value, 1),
                            function_class=state.functionClass,
                            function_instance=state.functionInstance,
                        )
                    )
                elif state.functionInstance == "cooling-target":
                    climate_data["target_temperature_cooling"] = generate_target_temp(
                        func_def["values"][0], state
                    )
                elif state.functionInstance == "heating-target":
                    climate_data["target_temperature_heating"] = generate_target_temp(
                        func_def["values"][0], state
                    )
                elif state.functionInstance == "auto-cooling-target":
                    climate_data["target_temperature_auto_cooling"] = (
                        generate_target_temp(func_def["values"][0], state)
                    )
                elif state.functionInstance == "auto-heating-target":
                    climate_data["target_temperature_auto_heating"] = (
                        generate_target_temp(func_def["values"][0], state)
                    )
                elif state.functionInstance == "safety-mode-max-temp":
                    climate_data["safety_max_temp"] = generate_target_temp(
                        func_def["values"][0], state
                    )
                elif state.functionInstance == "safety-mode-min-temp":
                    climate_data["safety_min_temp"] = generate_target_temp(
                        func_def["values"][0], state
                    )
            elif state.functionClass == "mode":
                all_modes = set(process_function(afero_device.functions, "mode"))
                climate_data["hvac_mode"] = features.HVACModeFeature(
                    mode=state.value,
                    previous_mode=state.value,
                    modes=all_modes,
                    supported_modes=all_modes,
                )
            elif state.functionClass == "available":
                climate_data["available"] = state.value
            elif number := await self.initialize_number(func_def, state):
                climate_data["numbers"][number[0]] = number[1]
            elif select := await self.initialize_select(afero_device.functions, state):
                climate_data["selects"][select[0]] = select[1]
            elif state.functionClass == "temperature-units":
                climate_data["display_celsius"] = state.value == "celsius"

        return climate_data

    async def update_climate_elem(self, afero_device: AferoDevice) -> set:
        """Update the climate elements of a device."""
        updated_keys = set()
        cur_item = self.get_device(afero_device.id)

        temperature_mapping = {
            "auto-heating-target": "target_temperature_auto_heating",
            "auto-cooling-target": "target_temperature_auto_cooling",
            "cooling-target": "target_temperature_cooling",
            "heating-target": "target_temperature_heating",
            "safety-mode-max-temp": "safety_max_temp",
            "safety-mode-min-temp": "safety_min_temp",
        }

        for state in afero_device.states:
            if state.functionClass == "available":
                if cur_item.available != state.value:
                    cur_item.available = state.value
                    updated_keys.add("available")
            elif state.functionClass == "temperature":
                if state.functionInstance == "current-temp":
                    if cur_item.current_temperature:
                        temp_value = cur_item.current_temperature.temperature
                        rounded_val = round(state.value, 1)
                        if temp_value != rounded_val:
                            cur_item.current_temperature.temperature = rounded_val
                            updated_keys.add(f"temperature-{state.functionInstance}")
                elif state.functionInstance in temperature_mapping:
                    temp_item = getattr(
                        cur_item, temperature_mapping.get(state.functionInstance), None
                    )
                    if temp_item and temp_item.value != state.value:
                        temp_item.value = state.value
                        updated_keys.add(f"temperature-{state.functionInstance}")
            elif state.functionClass == "mode":
                if cur_item.hvac_mode and cur_item.hvac_mode.mode != state.value:
                    # We only want to update the previous mode when we are in heat or cool
                    if cur_item.hvac_mode.mode in ["cool", "heat"]:
                        cur_item.hvac_mode.previous_mode = cur_item.hvac_mode.mode
                    cur_item.hvac_mode.mode = state.value
                    updated_keys.add(state.functionClass)
            elif (update_key := await self.update_number(state, cur_item)) or (
                update_key := await self.update_select(state, cur_item)
            ):
                updated_keys.add(update_key)
            elif state.functionClass == "temperature-units":
                new_val: bool = state.value == "celsius"
                if cur_item.display_celsius != new_val:
                    cur_item.display_celsius = new_val
                    updated_keys.add(state.functionClass)
        return updated_keys

    async def set_climate_state(self, device_id: str, update_obj, **kwargs) -> None:
        """Set climate state."""
        try:
            cur_item = self.get_device(device_id)
        except DeviceNotFound:
            self._logger.info("Unable to find device %s", device_id)
            return

        is_celsius = kwargs.get("is_celsius", False)

        temperature_kwargs = {
            "target_temperature_auto_heating": "target_temperature_auto_heating",
            "target_temperature_auto_cooling": "target_temperature_auto_cooling",
            "target_temperature_heating": "target_temperature_heating",
            "target_temperature_cooling": "target_temperature_cooling",
            "safety_min_temp": "safety_min_temp",
            "safety_max_temp": "safety_max_temp",
        }

        for kwarg_name, attr_name in temperature_kwargs.items():
            temp_val = kwargs.get(kwarg_name)
            if temp_val is not None:
                cur_temp_feature = getattr(cur_item, attr_name, None)
                if cur_temp_feature:
                    if not cur_item.display_celsius and not is_celsius:
                        temp_val = calculate_hubspace_celsius(temp_val)

                    setattr(
                        update_obj,
                        attr_name,
                        features.TargetTemperatureFeature(
                            value=temp_val,
                            min=cur_temp_feature.min,
                            max=cur_temp_feature.max,
                            step=cur_temp_feature.step,
                            instance=cur_temp_feature.instance,
                        ),
                    )

        await self.update(device_id, obj_in=update_obj)



"""Representation of an Afero Light and its corresponding updates."""

from dataclasses import dataclass
import re

from aioafero.v1.models import features

from .resource import ResourceTypes
from .standard_mixin import StandardMixin

rgbw_name_search = re.compile(r"RGB\w*W")


@dataclass(kw_only=True)
class Light(StandardMixin):
    """Representation of an Afero Light."""

    type: ResourceTypes = ResourceTypes.LIGHT

    on: features.OnFeature | None = None
    color: features.ColorFeature | None = None
    color_mode: features.ColorModeFeature | None = None
    color_modes: list[str] | None = None
    color_temperature: features.ColorTemperatureFeature | None = None
    dimming: features.DimmingFeature | None = None
    effect: features.EffectFeature | None = None
    supports_white: bool = False

    def __post_init__(self):
        """Determine if white only is supported."""
        super().__post_init__()
        model = getattr(getattr(self, "device_information", None), "model", None)
        if model is not None and rgbw_name_search.search(model):
            self.supports_white = True

    @property
    def supports_color(self) -> bool:
        """Return if this light supports color control."""
        return self.color is not None

    @property
    def supports_color_temperature(self) -> bool:
        """Return if this light supports color_temperature control."""
        return self.color_temperature is not None

    @property
    def supports_color_white(self) -> bool:
        """Return if this light supports setting white."""
        return self.color_modes is not None and "white" in self.color_modes

    @property
    def supports_dimming(self) -> bool:
        """Return if this light supports brightness control."""
        return self.dimming is not None

    @property
    def supports_effects(self) -> bool:
        """Return if this light supports brightness control."""
        return self.effect is not None

    @property
    def supports_on(self):
        """If the light can be toggled on or off."""
        return self.on is not None

    @property
    def is_on(self) -> bool:
        """Return bool if light is currently powered on."""
        if self.on is not None:
            return self.on.on
        return False

    @property
    def brightness(self) -> float:
        """Return current brightness of light."""
        if self.dimming is not None:
            return self.dimming.brightness
        return 100.0 if self.is_on else 0.0


@dataclass
class LightPut[AferoResource]:
    """States that can be updated for a light."""

    on: features.OnFeature | None = None
    color: features.ColorFeature | None = None
    color_mode: features.ColorModeFeature | None = None
    color_temperature: features.ColorTemperatureFeature | None = None
    dimming: features.DimmingFeature | None = None
    effect: features.EffectFeature | None = None



import pytest

from aioafero.v1.models import SecuritySystem, features, DeviceInformation


@pytest.fixture
def populated_entity():
    return SecuritySystem(
        _id="entity-1",
        available=True,
        alarm_state=features.ModeFeature(
            mode="arm-away", modes={"arm-away", "disarmed", "arm-stay", "alarming-sos"}
        ),
        numbers={
            ("arm-exit-delay", "away"): features.NumbersFeature(
                value=0,
                min=0,
                max=300,
                step=1,
                name="Exit Delay - Away",
                unit="seconds",
            ),
        },
        selects={
            ("volume ", "siren "): features.SelectFeature(
                selected="volume-04",
                selects={
                    "volume-00",
                    "volume-01",
                    "volume-02",
                    "volume-03",
                    "volume-04",
                },
                name="Siren Volume",
            ),
        },
        device_information=DeviceInformation(
            functions=[
            {
                "functionClass": "preset",
                "functionInstance": "preset-1",
                "value": "on",
                "lastUpdateTime": 0,
            }
        ]
        )
    )


@pytest.fixture
def empty_entity():
    return SecuritySystem(
        _id="entity-1",
        available=True,
        alarm_state=None,
        numbers={},
        selects={},
    )


def test_init(populated_entity):
    assert populated_entity.id == "entity-1"
    assert populated_entity.available is True
    assert populated_entity.instances == {"preset": "preset-1"}
    assert populated_entity.supports_away is True
    assert populated_entity.supports_arm_bypass is False
    assert populated_entity.supports_home is True
    assert populated_entity.supports_night is False
    assert populated_entity.supports_vacation is False
    assert populated_entity.supports_trigger is True


def test_init_empty(empty_entity):
    assert empty_entity.alarm_state is None


def test_get_instance(populated_entity):
    assert populated_entity.get_instance("preset") == "preset-1"



"""Handles generic information related to all HVAC devices."""

from abc import ABC, abstractmethod
from dataclasses import dataclass

from aioafero.util import calculate_hubspace_fahrenheit

from . import features


@dataclass(kw_only=True)
class HVACMixin(ABC):
    """Mixin for HVAC properties and methods."""

    display_celsius: bool | None = None
    current_temperature: features.CurrentTemperatureFeature | None = None
    fan_running: bool | None = None
    fan_mode: features.ModeFeature | None = None
    hvac_action: str | None = None
    hvac_mode: features.HVACModeFeature | None = None
    target_temperature_heating: features.TargetTemperatureFeature | None = None
    target_temperature_cooling: features.TargetTemperatureFeature | None = None
    target_temperature_auto_heating: features.TargetTemperatureFeature | None = None
    target_temperature_auto_cooling: features.TargetTemperatureFeature | None = None

    @property
    def target_temperature(self) -> float | None:
        """Temperature which the HVAC will try to achieve."""
        if self.hvac_mode is None or self.hvac_mode.mode not in [
            "cool",
            "heat",
            "fan",
            "off",
            "auto-cool",
        ]:
            return None
        target_feature = self._get_target_feature(self.get_mode_to_check())
        if not target_feature:
            return None
        celsius: float | None = getattr(target_feature, "value", None)
        if self.display_celsius:
            return celsius
        return calculate_hubspace_fahrenheit(celsius)

    @abstractmethod
    def get_mode_to_check(self) -> str | None:
        """Determine the current mode of the thermostat."""

    def _get_target_feature(
        self, mode: str
    ) -> features.TargetTemperatureFeature | None:
        if mode == "cool":
            return getattr(self, "target_temperature_cooling", None)
        if mode == "heat":
            return getattr(self, "target_temperature_heating", None)
        return None

    @property
    def target_temperature_step(self) -> float:
        """Smallest increment for adjusting the temperature."""
        set_mode = self.get_mode_to_check()
        if not set_mode:
            val = 0.5  # Default from Hubspace
        else:
            target_feature = self._get_target_feature(set_mode)
            val = getattr(target_feature, "step", 0.5)
        if self.display_celsius:
            return val
        return 1

    @property
    def target_temperature_max(self) -> float | None:
        """Maximum target temperature."""
        set_mode = self.get_mode_to_check()
        if not set_mode or self.hvac_mode.mode == "auto":
            auto_cooling = getattr(self, "target_temperature_auto_cooling", None)
            val = getattr(auto_cooling, "max", None)
        else:
            target_feature = self._get_target_feature(set_mode)
            val = getattr(target_feature, "max", None)

        if val is None:
            return None

        if self.display_celsius:
            return val
        return calculate_hubspace_fahrenheit(val)

    @property
    def target_temperature_min(self) -> float | None:
        """Minimum target temperature."""
        set_mode = self.get_mode_to_check()
        if not set_mode or self.hvac_mode.mode == "auto":
            auto_heating = getattr(self, "target_temperature_auto_heating", None)
            val = getattr(auto_heating, "min", None)
        else:
            target_feature = self._get_target_feature(set_mode)
            val = getattr(target_feature, "min", None)

        if val is None:
            return None

        if self.display_celsius:
            return val
        return calculate_hubspace_fahrenheit(val)

    @property
    def temperature(self) -> float | None:
        """Current temperature of the selected mode."""
        if self.current_temperature is None:
            return None
        if self.display_celsius:
            return self.current_temperature.temperature
        return calculate_hubspace_fahrenheit(self.current_temperature.temperature)



"""Test LightController"""

import pytest
from dataclasses import asdict

from aioafero.device import AferoState
from aioafero.v1.controllers import event, light
from aioafero.v1.controllers.light import LightController, features, process_color_temps
from aioafero.v1.models.features import EffectFeature
from aioafero.v1.models.light import Light
from aioafero.v1.models.resource import ResourceTypes, DeviceInformation

from .. import utils

a21_light = utils.create_devices_from_data("light-a21.json")[0]
zandra_light = utils.create_devices_from_data("fan-ZandraFan.json")[1]
dimmer_light = utils.create_devices_from_data("dimmer-HPDA1110NWBP.json")[0]
flushmount_light = utils.create_devices_from_data("light-flushmount.json")[0]
flushmount_light_color_id = f"{flushmount_light.id}-light-color"
flushmount_light_white_id = f"{flushmount_light.id}-light-white"

speaker_power_light = utils.create_devices_from_data("light-with-speaker.json")[0]
speaker_power_light_id = f"{speaker_power_light.id}-light-speaker-power"


@pytest.fixture
def mocked_controller(mocked_bridge, mocker):
    mocker.patch("time.time", return_value=12345)
    return mocked_bridge.lights


def test_generate_split_name():
    assert (
        light.generate_split_name(a21_light, "beans") == f"{a21_light.id}-light-beans"
    )


@pytest.mark.parametrize(
    "device, expected",
    [
        # Flushmount splits
        (
            flushmount_light,
            [
                ("color", ResourceTypes.LIGHT),
                ("white", ResourceTypes.LIGHT),
            ],
        ),
        # No splits
        (a21_light, []),
        (zandra_light, []),
        (dimmer_light, []),
    ],
)
def test_get_split_instances(device, expected):
    assert light.get_split_instances(device) == expected


@pytest.mark.parametrize(
    "device, instance, expected",
    [
        # Has no split instances
        (a21_light, "nada", []),
        # Flushmount white
        (
            flushmount_light,
            "white",
            [
                AferoState(
                    functionClass="toggle",
                    value="off",
                    lastUpdateTime=0,
                    functionInstance="white",
                ),
                AferoState(
                    functionClass="brightness",
                    value=100,
                    lastUpdateTime=0,
                    functionInstance="white",
                ),
                AferoState(
                    functionClass="available",
                    value=True,
                    lastUpdateTime=0,
                    functionInstance=None,
                ),
            ],
        ),
        # Flushmount color
        (
            flushmount_light,
            "color",
            [
                AferoState(
                    functionClass="brightness",
                    value=1,
                    lastUpdateTime=0,
                    functionInstance="color",
                ),
                AferoState(
                    functionClass="restore-values",
                    value="partial-restore",
                    lastUpdateTime=0,
                    functionInstance=None,
                ),
                AferoState(
                    functionClass="color-sequence",
                    value="sleep",
                    lastUpdateTime=0,
                    functionInstance="custom",
                ),
                AferoState(
                    functionClass="toggle",
                    value="on",
                    lastUpdateTime=0,
                    functionInstance="color",
                ),
                AferoState(
                    functionClass="color-mode",
                    value="color",
                    lastUpdateTime=0,
                    functionInstance=None,
                ),
                AferoState(
                    functionClass="color-rgb",
                    value={
                        "color-rgb": {
                            "b": 204,
                            "g": 242,
                            "r": 255,
                        },
                    },
                    lastUpdateTime=0,
                    functionInstance=None,
                ),
                AferoState(
                    functionClass="speed",
                    value=0,
                    lastUpdateTime=0,
                    functionInstance="color-sequence",
                ),
                AferoState(
                    functionClass="color-sequence",
                    value="custom",
                    lastUpdateTime=0,
                    functionInstance="preset",
                ),
                AferoState(
                    functionClass="color-temperature",
                    value=3000,
                    lastUpdateTime=0,
                    functionInstance=None,
                ),
                AferoState(
                    functionClass="wifi-ssid",
                    value="c87d78a4-3f6e-4468-a034-4bef7b7cd4b3",
                    lastUpdateTime=0,
                    functionInstance=None,
                ),
                AferoState(
                    functionClass="wifi-rssi",
                    value=-37,
                    lastUpdateTime=0,
                    functionInstance=None,
                ),
                AferoState(
                    functionClass="wifi-steady-state",
                    value="connected",
                    lastUpdateTime=0,
                    functionInstance=None,
                ),
                AferoState(
                    functionClass="wifi-setup-state",
                    value="connected",
                    lastUpdateTime=0,
                    functionInstance=None,
                ),
                AferoState(
                    functionClass="wifi-mac-address",
                    value="0ddc8684-2404-4e10-8495-fd5c82dda3e6",
                    lastUpdateTime=0,
                    functionInstance=None,
                ),
                AferoState(
                    functionClass="geo-coordinates",
                    value={
                        "geo-coordinates": {
                            "latitude": "0",
                            "longitude": "0",
                        },
                    },
                    lastUpdateTime=0,
                    functionInstance="system-device-location",
                ),
                AferoState(
                    functionClass="scheduler-flags",
                    value=1,
                    lastUpdateTime=0,
                    functionInstance=None,
                ),
                AferoState(
                    functionClass="available",
                    value=True,
                    lastUpdateTime=0,
                    functionInstance=None,
                ),
                AferoState(
                    functionClass="visible",
                    value=True,
                    lastUpdateTime=0,
                    functionInstance=None,
                ),
                AferoState(
                    functionClass="direct",
                    value=True,
                    lastUpdateTime=0,
                    functionInstance=None,
                ),
                AferoState(
                    functionClass="ble-mac-address",
                    value="444a8248-56ef-4735-a5a6-9dc486a16f85",
                    lastUpdateTime=0,
                    functionInstance=None,
                ),
            ],
        ),
        # Pull out speaker power
        (
            speaker_power_light,
            "speaker-power",
            [
                AferoState(
                    functionClass="toggle",
                    value="off",
                    lastUpdateTime=0,
                    functionInstance="speaker-power",
                ),
                AferoState(
                    functionClass="available",
                    value=True,
                    lastUpdateTime=0,
                    functionInstance=None,
                ),

            ]
        ),
    ],
)
def test_get_valid_states(device, instance, expected):
    assert light.get_valid_states(device, instance) == expected


def test_light_callback():
    multi_devs, remove_dev = light.light_callback(flushmount_light)
    assert remove_dev is True
    assert len(multi_devs) == 3
    assert len(multi_devs[0].states) == 20
    assert multi_devs[0].id == flushmount_light_color_id
    assert multi_devs[0].device_class == ResourceTypes.LIGHT.value
    assert len(multi_devs[1].states) == 3
    assert multi_devs[1].id == flushmount_light_white_id
    assert multi_devs[1].friendly_name == f"{flushmount_light.friendly_name} - white"
    assert multi_devs[1].device_class == ResourceTypes.LIGHT.value
    assert multi_devs[2].id == flushmount_light.id
    assert len(multi_devs[2].states) == 7
    assert multi_devs[2].device_class == "parent-device"


def test_light_speaker():
    multi_devs, remove_dev = light.light_callback(speaker_power_light)
    assert remove_dev is False
    assert len(multi_devs) == 1
    assert len(multi_devs[0].states) == 2
    assert multi_devs[0].id == speaker_power_light_id
    assert multi_devs[0].device_class == ResourceTypes.SWITCH.value


def test_light_callback_none():
    multi_devs, remove_dev = light.light_callback(a21_light)
    assert remove_dev is False
    assert len(multi_devs) == 0


@pytest.mark.asyncio
async def test_initialize_a21(mocked_controller):
    await mocked_controller.initialize_elem(a21_light)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == "dd883754-e9f2-4c48-b755-09bf6ce776be"
    assert dev.on == features.OnFeature(on=True)
    assert dev.color == features.ColorFeature(red=232, green=255, blue=30)
    assert dev.color_mode == features.ColorModeFeature(mode="white")
    assert dev.color_temperature == features.ColorTemperatureFeature(
        temperature=4000,
        supported=[
            2200,
            2300,
            2400,
            2500,
            2600,
            2700,
            2800,
            2900,
            3000,
            3100,
            3200,
            3300,
            3400,
            3500,
            3600,
            3700,
            3800,
            3900,
            4000,
            4100,
            4200,
            4300,
            4400,
            4500,
            4600,
            4700,
            4800,
            4900,
            5000,
            5100,
            5200,
            5300,
            5400,
            5500,
            5600,
            5700,
            5800,
            5900,
            6000,
            6100,
            6200,
            6300,
            6400,
            6500,
        ],
        prefix="",
    )
    assert dev.dimming == features.DimmingFeature(
        brightness=50,
        supported=[
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            11,
            12,
            13,
            14,
            15,
            16,
            17,
            18,
            19,
            20,
            21,
            22,
            23,
            24,
            25,
            26,
            27,
            28,
            29,
            30,
            31,
            32,
            33,
            34,
            35,
            36,
            37,
            38,
            39,
            40,
            41,
            42,
            43,
            44,
            45,
            46,
            47,
            48,
            49,
            50,
            51,
            52,
            53,
            54,
            55,
            56,
            57,
            58,
            59,
            60,
            61,
            62,
            63,
            64,
            65,
            66,
            67,
            68,
            69,
            70,
            71,
            72,
            73,
            74,
            75,
            76,
            77,
            78,
            79,
            80,
            81,
            82,
            83,
            84,
            85,
            86,
            87,
            88,
            89,
            90,
            91,
            92,
            93,
            94,
            95,
            96,
            97,
            98,
            99,
            100,
        ],
    )
    assert dev.effect == features.EffectFeature(
        effect="getting-ready",
        effects={
            "preset": {"jump-3", "fade-3", "fade-7", "jump-7", "flash"},
            "custom": {
                "dinner-party",
                "wake-up",
                "focus",
                "sleep",
                "valentines-day",
                "rainbow",
                "getting-ready",
                "christmas",
                "july-4th",
                "chill",
                "nightlight",
                "moonlight",
                "clarity",
            },
        },
    )


@pytest.mark.asyncio
async def test_initialize_zandra(mocked_controller):
    await mocked_controller.initialize_elem(zandra_light)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == "3a0c5015-c19d-417f-8e08-e71cd5bc221b"
    assert dev.on == features.OnFeature(
        on=True, func_class="power", func_instance="light-power"
    )
    assert dev.color is None
    assert dev.color_mode is None
    assert dev.color_temperature == features.ColorTemperatureFeature(
        temperature=3000, supported=[2700, 3000, 3500, 4000, 5000, 6500], prefix="K"
    )


@pytest.mark.asyncio
async def test_initialize_dimmer(mocked_controller):
    await mocked_controller.initialize_elem(dimmer_light)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == "ebda9f3b-05bc-4764-a9f7-e2d52f707130"
    assert dev.on == features.OnFeature(
        on=False, func_class="power", func_instance="gang-1"
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "afero_dev",
    [
        (a21_light),
        (zandra_light),
        (dimmer_light),
    ],
)
async def test_turn_on(afero_dev, mocked_controller):
    bridge = mocked_controller._bridge
    await bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(afero_dev)]
    )
    await bridge.async_block_until_done()
    await mocked_controller.initialize_elem(afero_dev)
    dev = mocked_controller.items[0]
    dev.on.on = False
    await mocked_controller.turn_on(afero_dev.id)
    await bridge.async_block_until_done()
    assert dev.is_on


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "afero_dev",
    [
        (a21_light),
        (zandra_light),
        (dimmer_light),
    ],
)
async def test_turn_off(afero_dev, mocked_controller):
    bridge = mocked_controller._bridge
    await bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(afero_dev)]
    )
    await bridge.async_block_until_done()
    dev = mocked_controller.items[0]
    dev.on.on = True
    await mocked_controller.turn_off(afero_dev.id)
    await bridge.async_block_until_done()
    assert not dev.is_on


@pytest.mark.asyncio
async def test_set_color_temperature(mocked_controller):
    bridge = mocked_controller._bridge
    await bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(a21_light)]
    )
    await bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    dev.on.on = False
    dev.color_temperature.temperature = 2700
    dev.color_mode.mode = "color"
    await mocked_controller.set_color_temperature(a21_light.id, 3475)
    await mocked_controller._bridge.async_block_until_done()
    assert dev.is_on
    assert dev.color_mode.mode == "white"
    assert dev.color_temperature.temperature == 3500


@pytest.mark.asyncio
async def test_set_brightness(mocked_controller):
    bridge = mocked_controller._bridge
    await bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(a21_light)]
    )
    await bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    dev.on.on = False
    dev.dimming.brightness = 50
    await mocked_controller.set_brightness(a21_light.id, 60)
    await mocked_controller._bridge.async_block_until_done()
    assert dev.is_on
    assert dev.dimming.brightness == 60


@pytest.mark.asyncio
async def test_set_brightness_split(mocked_controller, mocker):
    mocker.patch("time.time", return_value=12345)
    await mocked_controller._bridge.events.generate_events_from_data(
        utils.create_hs_raw_from_dump("light-flushmount.json")
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller._bridge.lights._items) == 2
    dev = mocked_controller[flushmount_light_white_id]
    assert dev.on.on is False
    # Split devices need their IDs correctly set
    json_resp = mocker.AsyncMock()
    json_resp.return_value = {"metadeviceId": flushmount_light.id, "values": [
        {
            "functionClass": "toggle",
            "functionInstance": "white",
            "value": "on",
            "lastUpdateTime": 0,
        },
        {
            "functionClass": "brightness",
            "functionInstance": "white",
            "value": 20,
            "lastUpdateTime": 0,
        }
    ]}
    resp = mocker.AsyncMock()
    resp.json = json_resp
    resp.status = 200
    mocker.patch.object(mocked_controller, "update_afero_api", return_value=resp)
    # Run the test
    await mocked_controller.set_brightness(flushmount_light_white_id, 20)
    await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller[flushmount_light_white_id]
    assert dev.is_on
    assert dev.dimming.brightness == 20

@pytest.mark.asyncio
async def test_set_rgb(mocked_controller):
    bridge = mocked_controller._bridge
    await bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(a21_light)]
    )
    await bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    dev.on.on = False
    dev.color_mode.mode = "white"
    dev.color.red = 100
    dev.color.green = 100
    dev.color.blue = 100
    await mocked_controller.set_rgb(a21_light.id, 0, 20, 40)
    await mocked_controller._bridge.async_block_until_done()
    assert dev.is_on
    assert dev.color_mode.mode == "color"
    assert dev.color.red == 0
    assert dev.color.green == 20
    assert dev.color.blue == 40


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "effect",
    [
        ("rainbow"),
        ("fade-7"),
    ],
)
async def test_set_effect(effect, mocked_controller):
    bridge = mocked_controller._bridge
    await bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(a21_light)]
    )
    await bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    dev.on.on = False
    dev.effect.effect = None
    await mocked_controller.set_effect(a21_light.id, effect)
    await mocked_controller._bridge.async_block_until_done()
    assert dev.is_on
    assert dev.color_mode.mode == "sequence"
    assert dev.effect.effect == effect


@pytest.mark.asyncio
async def test_update_elem(mocked_controller):
    bridge = mocked_controller._bridge
    await bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(a21_light)]
    )
    await bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    dev.available = False
    dev.on.on = False
    dev_update = utils.create_devices_from_data("light-a21.json")[0]
    new_states = [
        AferoState(
            functionClass="color-temperature", value="3000K", lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="brightness", value=40, lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="color-rgb", value={
                    "color-rgb": {
                        "r": 2,
                        "g": 3,
                        "b": 4,
                    }
                }, lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="power", value="on", lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="color-mode", value="color", lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="available", value=True, lastUpdateTime=0, functionInstance=None
        ),
    ]
    for state in new_states:
        utils.modify_state(dev_update, state)
    updates = await mocked_controller.update_elem(dev_update)
    dev = mocked_controller.items[0]
    assert dev.on.on is True
    assert dev.color_temperature.temperature == 3000
    assert dev.dimming.brightness == 40
    assert dev.color.red == 2
    assert dev.color.green == 3
    assert dev.color.blue == 4
    assert dev.color_mode.mode == "color"
    assert updates == {
        "on",
        "color_temperature",
        "dimming",
        "color",
        "color_mode",
        "available",
    }


@pytest.mark.asyncio
async def test_update_elem_no_updates(mocked_controller):
    bridge = mocked_controller._bridge
    await bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(a21_light)]
    )
    await bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    assert not await mocked_controller.update_elem(a21_light)


states_custom = [
    AferoState(
        functionClass="color-sequence", functionInstance="preset", lastUpdateTime=0, value="custom"
    ),
    AferoState(
        functionClass="color-sequence", functionInstance="custom", lastUpdateTime=0, value="rainbow"
    ),
]

states_preset = [
    AferoState(
        functionClass="color-sequence", functionInstance="preset", lastUpdateTime=0, value="fade-7"
    ),
    AferoState(
        functionClass="color-sequence", functionInstance="custom", lastUpdateTime=0, value="rainbow",
    ),
]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "new_states, expected",
    [
        (states_custom, "rainbow"),
        (states_preset, "fade-7"),
    ],
)
async def test_update_elem_effect(new_states, expected, mocked_controller):
    bridge = mocked_controller._bridge
    await bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(a21_light)]
    )
    await bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev_update = utils.create_devices_from_data("light-a21.json")[0]
    for state in new_states:
        utils.modify_state(dev_update, state)
    await mocked_controller.update_elem(dev_update)
    dev = mocked_controller.items[0]
    assert dev.effect.effect == expected


@pytest.mark.asyncio
async def test_set_state_empty(mocked_controller):
    await mocked_controller.initialize_elem(a21_light)
    await mocked_controller.set_state(a21_light.id)


@pytest.mark.asyncio
async def test_light_emitting(bridge):
    dev_update = utils.create_devices_from_data("light-a21.json")[0]
    add_event = {
        "type": "add",
        "device_id": dev_update.id,
        "device": dev_update,
    }
    # Simulate a poll
    bridge.events.emit(event.EventType.RESOURCE_ADDED, add_event)
    await bridge.async_block_until_done()
    assert len(bridge.lights._items) == 1
    # Simulate an update
    utils.modify_state(
        dev_update,
        AferoState(
            functionClass="available",
            functionInstance=None,
            value=False,
        ),
    )
    update_event = {
        "type": "update",
        "device_id": dev_update.id,
        "device": dev_update,
    }
    bridge.events.emit(event.EventType.RESOURCE_UPDATED, update_event)
    await bridge.async_block_until_done()
    assert len(bridge.lights._items) == 1
    assert not bridge.lights._items[dev_update.id].available


@pytest.mark.asyncio
async def test_set_state_no_dev(mocked_controller, caplog):
    caplog.set_level(0)
    await mocked_controller.initialize_elem(a21_light)
    mocked_controller._bridge.add_device(a21_light.id, mocked_controller)
    await mocked_controller.set_state("not-a-device")
    mocked_controller._bridge.request.assert_not_called()
    assert "Unable to find device" in caplog.text


seq_custom = {
    "preset": AferoState(
        functionClass="color-sequence", value="custom", lastUpdateTime=0, functionInstance="preset"
    ),
    "custom": AferoState(
        functionClass="color-sequence", value="rainbow", lastUpdateTime=0, functionInstance="custom"
    ),
}

seq_preset = {
    "preset": AferoState(
        functionClass="color-sequence", value="fade-3", lastUpdateTime=0, functionInstance="preset"
    ),
    "custom": AferoState(
        functionClass="color-sequence", value="rainbow", lastUpdateTime=0, functionInstance="custom"
    ),
}

light1_effects = {
    "preset": {"fade-3"},
    "custom": {"rainbow"},
}
light1 = Light(
    _id="test-light-1",
    available=True,
    effect=EffectFeature(effect="getting-ready", effects=light1_effects),
    device_information=DeviceInformation(model="TBD"),
)
light1_no_update = Light(
    _id="test-light-1",
    available=True,
    effect=EffectFeature(effect="rainbow", effects=light1_effects),
device_information=DeviceInformation(model="TBD"),
)
light1_no_update_preset = Light(
    _id="test-light-1",
    available=True,
    effect=EffectFeature(effect="fade-3", effects=light1_effects),
device_information=DeviceInformation(model="TBD"),
)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "elem, color_seq_states, expected_effect, updated",
    [
        (light1, {}, "getting-ready", False),
        (light1, seq_custom, "rainbow", True),
        (light1_no_update, seq_custom, "rainbow", False),
        (light1, seq_preset, "fade-3", True),
        (light1_no_update_preset, seq_preset, "fade-3", False),
    ],
)
async def test_update_elem_color(
    mocked_controller, elem, color_seq_states, expected_effect, updated
):
    updates = await mocked_controller.update_elem_color(elem, color_seq_states)
    assert len(updates) == int(updated)
    assert elem.effect.effect == expected_effect


def test_process_color_temps():
    temps = [{"name": "2700K"}, {"name": "3000"}]
    assert process_color_temps(temps) == [2700, 3000]


@pytest.mark.asyncio
async def test_emitting(bridge):
    # Simulate the discovery process
    await bridge.events.generate_events_from_data(
        utils.create_hs_raw_from_dump("light-flushmount.json")
    )
    await bridge.async_block_until_done()
    assert len(bridge.lights._items) == 2
    assert bridge.lights[flushmount_light_color_id].on.on
    assert bridge.lights[flushmount_light_color_id].brightness == 1
    assert not bridge.lights[flushmount_light_white_id].on.on
    assert bridge.lights[flushmount_light_white_id].brightness == 100
    assert bridge.devices[flushmount_light.id].available is True
    dev_update = utils.create_devices_from_data("light-flushmount.json")[0]
    # Simulate an update
    utils.modify_state(
        dev_update,
        AferoState(
            functionClass="toggle",
            functionInstance="white",
            value="on",
        ),
    )
    utils.modify_state(
        dev_update,
        AferoState(
            functionClass="brightness",
            functionInstance="white",
            value=50,
        ),
    )
    utils.modify_state(
        dev_update,
        AferoState(
            functionClass="toggle",
            functionInstance="color",
            value="off",
        ),
    )
    utils.modify_state(
        dev_update,
        AferoState(
            functionClass="brightness",
            functionInstance="color",
            value=55,
        ),
    )
    utils.modify_state(
        dev_update,
        AferoState(
            functionClass="available",
            functionInstance=None,
            value=False,
        ),
    )
    await bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(dev_update)]
    )
    await bridge.async_block_until_done()
    assert bridge.lights[flushmount_light_color_id].brightness == 55
    assert not bridge.lights[flushmount_light_color_id].on.on
    assert bridge.lights[flushmount_light_white_id].brightness == 50
    assert bridge.lights[flushmount_light_white_id].on.on
    assert bridge.devices[flushmount_light.id].available is False


@pytest.mark.asyncio
async def test_set_state_white_light(mocked_controller):
    bridge = mocked_controller._bridge
    await bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(speaker_power_light)]
    )
    await bridge.async_block_until_done()
    await mocked_controller.set_state(speaker_power_light.id, on=True, force_white_mode=75)
    await mocked_controller._bridge.async_block_until_done()
    assert mocked_controller[speaker_power_light.id].on.on is True
    assert mocked_controller[speaker_power_light.id].color_mode.mode == "white"
    assert mocked_controller[speaker_power_light.id].dimming.brightness == 75



"""Controls Hubspace devices on v1 API."""

__all__ = [
    "AferoBridgeV1",
    "AferoController",
    "AferoModelResource",
    "BaseResourcesController",
    "DeviceController",
    "FanController",
    "LightController",
    "LockController",
    "PortableACController",
    "SecuritySystemController",
    "SecuritySystemKeypadController",
    "SecuritySystemSensorController",
    "SwitchController",
    "ThermostatController",
    "TokenData",
    "ValveController",
    "models",
]

import asyncio
from collections.abc import Callable, Generator
import contextlib
from contextlib import asynccontextmanager
import logging
from typing import Any

import aiohttp
from aiohttp import web_exceptions
from securelogging import LogRedactorMessage, add_secret

from aioafero.device import AferoDevice, AferoResource, AferoState
from aioafero.errors import (
    AferoError,
    DeviceNotFound,
    ExceededMaximumRetries,
    InvalidAuth,
)

from . import models, v1_const
from .auth import AferoAuth, TokenData, passthrough
from .controllers.base import AferoBinarySensor, AferoSensor, BaseResourcesController
from .controllers.device import DeviceController
from .controllers.event import EventCallBackType, EventStream, EventType
from .controllers.exhaust_fan import ExhaustFanController
from .controllers.fan import FanController
from .controllers.light import LightController
from .controllers.lock import LockController
from .controllers.portable_ac import PortableACController
from .controllers.security_system import SecuritySystemController
from .controllers.security_system_keypad import SecuritySystemKeypadController
from .controllers.security_system_sensor import SecuritySystemSensorController
from .controllers.switch import SwitchController
from .controllers.thermostat import ThermostatController
from .controllers.valve import ValveController

type AferoModelResource = (
    models.Device
    | models.Fan
    | models.Light
    | models.Lock
    | models.Switch
    | models.Valve
    | models.Thermostat
    | AferoBinarySensor
    | AferoSensor
    | models.ExhaustFan
    | models.PortableAC
    | models.SecuritySystem
    | models.SecuritySystemSensor
)

type AferoController = (
    DeviceController
    | FanController
    | LightController
    | LockController
    | AferoSensor
    | SwitchController
    | ThermostatController
    | ValveController
    | ExhaustFanController
    | PortableACController
    | SecuritySystemController
    | SecuritySystemKeypadController
    | SecuritySystemSensorController
)


class AferoBridgeV1:
    """Controls Afero IoT devices on v1 API."""

    _web_session: aiohttp.ClientSession | None = None

    def __init__(
        self,
        username: str,
        password: str,
        refresh_token: str | None = None,
        session: aiohttp.ClientSession | None = None,
        polling_interval: int = 30,
        afero_client: str | None = "hubspace",
        hide_secrets: bool = True,
        poll_version: bool = True,
        client_name: str | None = "aioafero",
    ):
        """Create a bridge that communicates with Afero IoT API."""
        if hide_secrets:
            self.secret_logger = LogRedactorMessage
        else:
            self.secret_logger = passthrough
        self._close_session: bool = session is None
        self._web_session: aiohttp.ClientSession = session
        self._account_id: str | None = None
        self._afero_client: str = afero_client
        self._auth = AferoAuth(
            self,
            username,
            password,
            refresh_token=refresh_token,
            afero_client=afero_client,
            hide_secrets=hide_secrets,
        )
        self.client_name = client_name
        self.logger = logging.getLogger(f"{__package__}-{afero_client}[{username}]")
        if len(self.logger.handlers) == 0:
            self.logger.addHandler(logging.StreamHandler())
        self._known_devs: dict[str, BaseResourcesController] = {}
        self._known_afero_devices: dict[str, str] = {}
        # Known running tasks
        self._scheduled_tasks: list[asyncio.Task] = []
        self._adhoc_tasks: list[asyncio.Task] = []
        # Data Updater
        self._events: EventStream = EventStream(self, polling_interval, poll_version)
        # Data Controllers
        self._controllers: dict[str, BaseResourcesController] = {}
        self.add_controller("devices", DeviceController)
        self.add_controller("exhaust_fans", ExhaustFanController)
        self.add_controller("fans", FanController)
        self.add_controller("lights", LightController)
        self.add_controller("locks", LockController)
        self.add_controller("portable_acs", PortableACController)
        self.add_controller("security_systems", SecuritySystemController)
        self.add_controller("security_systems_keypads", SecuritySystemKeypadController)
        self.add_controller("security_systems_sensors", SecuritySystemSensorController)
        self.add_controller("switches", SwitchController)
        self.add_controller("thermostats", ThermostatController)
        self.add_controller("valves", ValveController)

    @property
    def refresh_token(self) -> str | None:
        """Get the current sessions refresh token."""
        return self._auth.refresh_token

    @property
    def events(self) -> EventStream:
        """Get the class that handles getting new data and notifying controllers."""
        return self._events

    @property
    def controllers(self) -> list:
        """Get a list of initialized controllers."""
        return [
            controller
            for controller in self._controllers.values()
            if controller.initialized
        ]

    @property
    def tracked_devices(self) -> set:
        """Get all tracked devices."""
        return set(self._known_devs.keys())

    async def otp_login(self, otp_code: str) -> None:
        """Perform OTP login with the provided code."""
        await self._auth.perform_otp_login(otp_code)

    def add_device(
        self, device_id: str, controller: BaseResourcesController[AferoResource]
    ) -> None:
        """Add a device to the list of known devices."""
        self._known_devs[device_id] = controller

    def get_device_controller(self, device_id: str) -> BaseResourcesController:
        """Get the controller for a given device."""
        try:
            return self._known_devs[device_id]
        except KeyError as err:
            raise DeviceNotFound(f"Unable to find device {device_id}") from err

    def remove_device(self, device_id: str) -> None:
        """Remove a device from the list of known devices."""
        with contextlib.suppress(KeyError):
            self._known_devs.pop(device_id)
        with contextlib.suppress(KeyError):
            self._known_afero_devices.pop(device_id)

    def add_afero_dev(self, device: AferoDevice, device_id: str | None = None) -> None:
        """Add a tracked afero device."""
        if not device_id:
            device_id = device.id
        self._known_afero_devices[device_id] = device

    def get_afero_device(self, device_id: str) -> AferoDevice | None:
        """Get the afero device for a given id."""
        try:
            return self._known_afero_devices[device_id]
        except KeyError as err:
            raise DeviceNotFound(f"Unable to find device for {device_id}") from err

    @property
    def account_id(self) -> str:
        """Get the account ID for the Afero IoT account."""
        return self._account_id

    @property
    def afero_client(self) -> str:
        """Get identifier for Afero system."""
        return self._afero_client

    def add_controller(self, name: str, controller_type: type) -> None:
        """Add a controller to the list of controllers."""
        self._controllers[name] = controller_type(self)
        setattr(self, name, self._controllers[name])

    def set_token_data(self, data: TokenData) -> None:
        """Set TokenData used for querying the API."""
        self._auth.set_token_data(data)

    def set_polling_interval(self, polling_interval: int) -> None:
        """Set the time between polling Afero API."""
        self._events.polling_interval = polling_interval

    def generate_api_url(self, endpoint: str) -> str:
        """Generate a URL for the Afero API."""
        endpoint = endpoint.removeprefix("/")
        return f"https://{v1_const.AFERO_CLIENTS[self._afero_client]['API_HOST']}/{endpoint}"

    async def close(self) -> None:
        """Close connection and cleanup."""
        for task in self._scheduled_tasks:
            task.cancel()
            await task
        self._scheduled_tasks = []
        await self.events.stop()
        if self._close_session and self._web_session:
            await self._web_session.close()
        self.logger.info("Connection to bridge closed.")

    def subscribe(
        self,
        callback: EventCallBackType,
    ) -> Callable:
        """Subscribe to status changes for all resources.

        Returns:
            function to unsubscribe.

        """
        unsubscribes = [
            controller.subscribe(callback) for controller in self.controllers
        ]

        def unsubscribe():
            for unsub in unsubscribes:
                unsub()

        return unsubscribe

    async def get_account_id(self) -> str:
        """Lookup the account ID associated with the login."""
        if not self._account_id:
            self.logger.debug("Querying API for account id")
            headers = {"host": v1_const.AFERO_CLIENTS[self._afero_client]["API_HOST"]}
            url = self.generate_api_url(v1_const.AFERO_GENERICS["ACCOUNT_ID_ENDPOINT"])
            with self.secret_logger():
                self.logger.debug(
                    "GETURL: %s, Headers: %s",
                    url,
                    headers,
                )
            res = await self.request(
                "GET",
                url,
                headers=headers,
            )
            res.raise_for_status()
            json_data = await res.json()
            if len(json_data) == 0 or len(json_data.get("accountAccess", [])) == 0:
                raise AferoError("No account ID found")
            self._account_id = (
                json_data.get("accountAccess")[0].get("account").get("accountId")
            )
            add_secret(self._account_id)
        return self._account_id

    async def initialize(self) -> None:
        """Initialize the bridge for communication with Afero API."""
        if len(self._scheduled_tasks) == 0:
            await self.initialize_cleanup()
            await self.get_account_id()
            await asyncio.gather(
                *[
                    controller.initialize()
                    for controller in self._controllers.values()
                    if not controller.initialized
                ]
            )
            await self._events.initialize()

    async def fetch_data(self, version_poll=False) -> list[dict[Any, str]]:
        """Query the API."""
        self.logger.debug("Querying API for all data")
        headers = {
            "host": v1_const.AFERO_CLIENTS[self._afero_client]["API_DATA_HOST"],
        }
        params = {"expansions": "state,capabilities,semantics"}
        url = self.generate_api_url(
            v1_const.AFERO_GENERICS["API_DEVICE_ENDPOINT"].format(self.account_id)
        )
        res = await self.request(
            "get",
            url,
            headers=headers,
            params=params,
        )
        res.raise_for_status()
        data = await res.json()
        if not isinstance(data, list):
            raise TypeError(data)
        if version_poll:
            devs = {}
            for dev in data:
                if dev.get("typeId") != "metadevice.device":
                    continue
                dev_id = dev.get("deviceId")
                if dev_id in devs:
                    dev["version_data"] = devs[dev_id]
                    continue
                dev["version_data"] = await self.get_device_version(dev_id)
                devs[dev_id] = dev["version_data"]

        return data

    async def fetch_device_states(self, device_id) -> list[dict[Any, str]]:
        """Query the API for new device states."""
        self.logger.debug("Querying the API for updated states for %s", device_id)
        headers = {
            "host": v1_const.AFERO_CLIENTS[self._afero_client]["API_DATA_HOST"],
        }
        params = {"expansions": "state,capabilities,semantics"}
        url = self.generate_api_url(
            v1_const.AFERO_GENERICS["API_DEVICE_STATE_ENDPOINT"].format(
                self.account_id, device_id
            )
        )
        res = await self.request(
            "get",
            url,
            headers=headers,
            params=params,
        )
        res.raise_for_status()
        data = await res.json()
        states = []
        for state in data.get("values", []):
            try:
                states.append(AferoState(**state))
            except TypeError:
                continue
        return states

    async def get_device_version(self, device_id: str) -> dict:
        """Query the API for device version information."""
        endpoint = v1_const.AFERO_GENERICS["API_DEVICE_VERSIONS_ENDPOINT"].format(
            self.account_id, device_id
        )
        url = self.generate_api_url(endpoint)
        res = await self.request("GET", url)
        res.raise_for_status()
        return await res.json()

    @asynccontextmanager
    async def create_request(
        self, method: str, url: str, include_token: bool, **kwargs
    ) -> Generator[aiohttp.ClientResponse, None, None]:
        """Make a request to any path with V2 request method (auth in header).

        Returns a generator with aiohttp ClientResponse.
        """
        if self._web_session is None:
            connector = aiohttp.TCPConnector(
                limit_per_host=3,
            )
            self._web_session = aiohttp.ClientSession(connector=connector)

        extras = {}
        if include_token:
            try:
                extras["Authorization"] = f"Bearer {await self._auth.token()}"
            except InvalidAuth:
                self.events.emit(EventType.INVALID_AUTH)
                raise
        headers = self.get_headers(**extras)
        headers.update(kwargs.get("headers", {}))
        kwargs["headers"] = headers
        kwargs["ssl"] = True
        async with self._web_session.request(method, url, **kwargs) as res:
            yield res

    async def request(
        self, method: str, url: str, include_token: bool = True, **kwargs
    ) -> aiohttp.ClientResponse:
        """Make request on the api and return response data."""
        retries = 0
        with self.secret_logger():
            self.logger.info("Making request [%s] to %s with %s", method, url, kwargs)
        while retries < v1_const.MAX_RETRIES:
            retries += 1
            if retries > 1:
                retry_wait = 0.25 * retries
                await asyncio.sleep(retry_wait)
            async with self.create_request(
                method, url, include_token, **kwargs
            ) as resp:
                # 504 means the API is overloaded, back off a bit
                # 503 means the service is temporarily unavailable, back off a bit.
                # 429 means the bridge is rate limiting/overloaded, we should back off a bit.
                if resp.status in [429, 503, 504]:
                    continue
                # 403 is bad auth
                if resp.status == 403:
                    raise web_exceptions.HTTPForbidden
                await resp.read()
                return resp
        raise ExceededMaximumRetries("Exceeded maximum number of retries")

    async def send_service_request(self, device_id: str, states: list[dict[str, Any]]):
        """Manually send state requests to Afero IoT.

        :param device_id: ID for the device
        :param states: List of states to send
        """
        controller = self._known_devs.get(device_id)
        if not controller:
            raise DeviceNotFound(f"Unable to find device {device_id}")
        await controller.update(device_id, states=states)

    def get_headers(self, **kwargs):
        """Get default headers for an API call."""
        headers: dict[str, str] = {
            "user-agent": v1_const.AFERO_GENERICS["DEFAULT_USERAGENT"].safe_substitute(
                client_name=self.client_name
            ),
            "accept-encoding": "gzip",
        }
        headers.update(kwargs)
        return headers

    # Task management enables us to block until finished
    def add_job(self, task: asyncio.Task) -> None:
        """Add a job to be processed."""
        self._adhoc_tasks.append(task)

    async def async_block_until_done(self):
        """Sync call for ensuring all processing is done."""
        await asyncio.gather(*self._adhoc_tasks)
        await self.events.async_block_until_done()

    async def initialize_cleanup(self) -> None:
        """Create the job that removes finished tasks."""
        self._scheduled_tasks.append(asyncio.create_task(self.__cleanup_processor()))

    async def __cleanup_processor(self) -> None:
        """Remove finished tasks."""
        with contextlib.suppress(asyncio.CancelledError):
            while True:
                for task in self._adhoc_tasks[:]:
                    if task.done():
                        self._adhoc_tasks.remove(task)
                await asyncio.sleep(1)



import pytest

from aioafero.v1.models import features, DeviceInformation
from aioafero.v1.models.light import Light


@pytest.fixture
def populated_light():
    return Light(
        _id="entity-1",
        available=True,
        on=features.OnFeature(on=True),
        color=features.ColorFeature(red=10, green=20, blue=40),
        color_mode=features.ColorModeFeature(mode="white"),
        color_modes=["white", "color"],
        color_temperature=features.ColorTemperatureFeature(
            temperature=3000, supported=list(range(2700, 5000, 100)), prefix="K"
        ),
        dimming=features.DimmingFeature(
            brightness=100, supported=list(range(0, 101, 10))
        ),
        effect=features.EffectFeature(
            effect="rainbow", effects={"custom": {"rainbow"}}
        ),
        device_information=DeviceInformation(
            model="AL-TP-RGBICTW-1",
            functions=[
            {
                "functionClass": "preset",
                "functionInstance": "preset-1",
                "value": "on",
                "lastUpdateTime": 0,
            }]
        )
    )


@pytest.fixture
def empty_light():
    return Light(
        _id="entity-1",
        available=True,
        on=None,
        color=None,
        color_mode=None,
        color_modes=[],
        color_temperature=None,
        dimming=None,
        effect=None,
        device_information=DeviceInformation(
            model="a21-light",
            functions=[
            {
                "functionClass": "preset",
                "functionInstance": "preset-1",
                "value": "on",
                "lastUpdateTime": 0,
            }]
        )
    )


def test_init(populated_light):
    assert populated_light.id == "entity-1"
    assert populated_light.available is True
    assert populated_light.on.on is True
    assert populated_light.color.red == 10
    assert populated_light.color_mode.mode == "white"
    assert populated_light.color_temperature.temperature == 3000
    assert populated_light.dimming.brightness == 100
    assert populated_light.effect.effect == "rainbow"
    assert populated_light.instances == {"preset": "preset-1"}
    assert populated_light.supports_on
    assert populated_light.supports_color
    assert populated_light.supports_color_temperature
    assert populated_light.supports_color_white
    assert populated_light.supports_dimming
    assert populated_light.supports_effects
    assert populated_light.is_on is True
    populated_light.on = None
    assert not populated_light.supports_on
    assert populated_light.brightness == 100
    assert populated_light.update_id == "entity-1"
    assert populated_light.instance is None
    populated_light._id = "entity-beans-1"
    populated_light.split_identifier = "beans"
    assert populated_light.update_id == "entity"
    assert populated_light.instance == "1"


def test_empty_light(empty_light):
    assert not empty_light.supports_on
    assert not empty_light.supports_color
    assert not empty_light.supports_color_temperature
    assert not empty_light.supports_color_white
    assert not empty_light.supports_dimming
    assert not empty_light.supports_effects
    assert not empty_light.is_on
    assert not empty_light.brightness
    assert not empty_light.supports_white


def test_get_instance(populated_light):
    assert populated_light.get_instance("preset") == "preset-1"



"""Errors used through aioafero."""


class AferoError(Exception):
    """Generic exception for Afero API or the responses."""


class DeviceNotFound(AferoError):
    """Device not found within the controller."""


class DeviceUpdateError(AferoError):
    """Unable to send a device update to Afero API."""


class ExceededMaximumRetries(AferoError):
    """Maximum retries exceeded when contacting Afero API."""


class InvalidAuth(AferoError):
    """Invalid credentials supplied during authentication."""


class InvalidResponse(AferoError):
    """An invalid response was received from Afero API."""


class SecuritySystemError(AferoError):
    """Generic exception for Afero Security System issues."""


class OTPError(AferoError):
    """Issue occurred with OTP authentication."""


class OTPRequired(AferoError):
    """One-time password is required for login."""


class InvalidOTP(OTPError):
    """One-time password provided is invalid."""



"""Test Portable ACs"""

import pytest

from aioafero.device import AferoState
from aioafero.v1.controllers.portable_ac import (
    PortableACController,
    features,
    generate_split_name,
    get_valid_states,
    portable_ac_callback,
)
from aioafero.v1.models import ResourceTypes

from .. import utils

portable_ac = utils.create_devices_from_data("portable-ac.json")[0]
portable_ac_id = "8d0414d6-a7f7-4bdb-99d5-d866318ff559"

portable_ac_swing = utils.create_devices_from_data("myko-portable-ac-with-swing.json")[1]
portable_ac_swing_id = "c31e1854-87de-47a7-ac62-f1cefe3ecba4"


@pytest.fixture
def mocked_controller(mocked_bridge, mocker):
    mocker.patch("time.time", return_value=12345)
    return mocked_bridge.portable_acs


def test_generate_split_name():
    assert (
        generate_split_name(portable_ac, "power")
        == "8d0414d6-a7f7-4bdb-99d5-d866318ff559-portable-ac-power"
    )


def test_get_valid_states():
    assert get_valid_states(portable_ac) == [
        AferoState(
            functionClass="power",
            value="off",
            lastUpdateTime=0,
            functionInstance=None,
        ),
        AferoState(
            functionClass="available",
            value=True,
            lastUpdateTime=0,
            functionInstance=None,
        ),
    ]


def test_exhaust_fan_callback():
    devs, remove_original = portable_ac_callback(portable_ac)
    assert remove_original is False
    assert len(devs) == 1
    expected_ids = [
        "8d0414d6-a7f7-4bdb-99d5-d866318ff559-portable-ac-power",
    ]
    for ind, dev in enumerate(devs):
        assert dev.id == expected_ids[ind]
        assert len(dev.states) == 2
        assert dev.device_class == ResourceTypes.SWITCH.value


def test_exhaust_fan_callback_wrong_dev():
    devs, remove_original = portable_ac_callback(
        utils.create_devices_from_data("light-a21.json")[0]
    )
    assert remove_original is False
    assert len(devs) == 0


@pytest.mark.asyncio
async def test_initialize(mocked_controller):
    await mocked_controller.initialize_elem(portable_ac)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == portable_ac_id
    assert dev.available is True
    assert dev.current_temperature == features.CurrentTemperatureFeature(
        temperature=35,
        function_class="temperature",
        function_instance="current-temp",
    )
    assert dev.hvac_mode == features.HVACModeFeature(
        mode="auto-cool",
        previous_mode="auto-cool",
        modes={"fan", "auto-cool", "dehumidify", "cool"},
        supported_modes={"fan", "auto-cool", "dehumidify", "cool"},
    )
    assert dev.target_temperature_cooling == features.TargetTemperatureFeature(
        value=22, step=0.5, min=16, max=30, instance="cooling-target"
    )
    assert dev.numbers == {}
    assert dev.selects == {
        ("fan-speed", "ac-fan-speed"): features.SelectFeature(
            selected="fan-speed-auto",
            selects={"fan-speed-auto", "fan-speed-2-100", "fan-speed-2-050"},
            name="Fan Speed",
        ),
        ("sleep", None): features.SelectFeature(
            selected="off",
            selects={"on", "off"},
            name="Sleep Mode",
        ),
    }
    assert dev.target_temperature_heating is None
    assert dev.target_temperature_auto_heating is None
    assert dev.target_temperature_auto_cooling is None


@pytest.mark.asyncio
async def test_initialize_swing(mocked_controller):
    await mocked_controller.initialize_elem(portable_ac_swing)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == portable_ac_swing_id
    assert dev.available is True
    assert dev.current_temperature == features.CurrentTemperatureFeature(
        temperature=21,
        function_class="temperature",
        function_instance="current-temp",
    )
    assert dev.hvac_mode == features.HVACModeFeature(
        mode="cool",
        previous_mode="cool",
        modes={"fan", "dehumidify", "heat", "cool"},
        supported_modes={"fan", "dehumidify", "heat", "cool"},
    )
    assert dev.target_temperature_cooling == features.TargetTemperatureFeature(
        value=21, step=1, min=16, max=31, instance="cooling-target"
    )
    assert dev.target_temperature_heating == features.TargetTemperatureFeature(
        value=31, step=1, min=16, max=31, instance="heating-target"
    )
    assert dev.numbers == {}
    assert dev.selects == {
        ("fan-speed", "ac-fan-speed"): features.SelectFeature(
            selected="fan-speed-3-033",
            selects={"fan-speed-3-100", "fan-speed-3-033", "fan-speed-3-066"},
            name="Fan Speed",
        ),
        ("sleep", None): features.SelectFeature(
            selected="off",
            selects={"on", "off"},
            name="Sleep Mode",
        ),
        ("air-swing", None): features.SelectFeature(
            selected="on",
            selects={"on", "off"},
            name="Swing",
        ),
    }


@pytest.mark.asyncio
async def test_update_elem(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(portable_ac)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev_update = utils.create_devices_from_data("portable-ac.json")[0]
    new_states = [
        AferoState(
            functionClass="available", value=False, lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="temperature", value=19, lastUpdateTime=0, functionInstance="current-temp"
        ),
        AferoState(
            functionClass="temperature", value=18, lastUpdateTime=0, functionInstance="cooling-target"
        ),
        AferoState(
            functionClass="mode", value="cool", lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="fan-speed",
            functionInstance="ac-fan-speed",
            lastUpdateTime=0,
            value="fan-speed-2-100",
        ),
        AferoState(
            functionClass="temperature-units",
            functionInstance=None,
            lastUpdateTime=0,
            value="celsius",
        ),
    ]
    for state in new_states:
        utils.modify_state(dev_update, state)
    updates = await mocked_controller.update_elem(dev_update)
    dev = mocked_controller.items[0]
    assert dev.available is False
    assert dev.current_temperature == features.CurrentTemperatureFeature(
        temperature=19,
        function_class="temperature",
        function_instance="current-temp",
    )
    assert dev.target_temperature_cooling.value == 18
    assert dev.hvac_mode.mode == "cool"
    assert dev.hvac_mode.previous_mode == "auto-cool"
    assert dev.selects[("fan-speed", "ac-fan-speed")].selected == "fan-speed-2-100"
    assert updates == {
        "available",
        "temperature-cooling-target",
        "temperature-current-temp",
        "mode",
        "select-('fan-speed', 'ac-fan-speed')",
        "temperature-units",
    }


@pytest.mark.asyncio
async def test_update_elem_no_updates(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(portable_ac)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev_update = utils.create_devices_from_data("portable-ac.json")[0]
    updates = await mocked_controller.update_elem(dev_update)
    assert updates == set()


@pytest.mark.asyncio
async def test_set_state(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(portable_ac)]
    )
    await mocked_controller._bridge.async_block_until_done()
    mocked_controller[portable_ac_id].display_celsius = True
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(
        portable_ac_id,
        hvac_mode="cool",
        target_temperature=22.5,
        selects={("fan-speed", "ac-fan-speed"): "fan-speed-2-100"},
    )
    await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller.items[0]
    assert dev.target_temperature_cooling.value == 22.5
    assert dev.hvac_mode.mode == "cool"
    assert dev.hvac_mode.previous_mode == "auto-cool"
    assert dev.selects[("fan-speed", "ac-fan-speed")].selected == "fan-speed-2-100"


@pytest.mark.asyncio
async def test_set_state_in_f(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(portable_ac)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(
        portable_ac_id,
        hvac_mode="cool",
        target_temperature=76,
        selects={("fan-speed", "ac-fan-speed"): "fan-speed-2-100"},
    )
    await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller.items[0]
    assert dev.target_temperature_cooling.value == 24.5
    assert dev.hvac_mode.mode == "cool"
    assert dev.hvac_mode.previous_mode == "auto-cool"
    assert dev.selects[("fan-speed", "ac-fan-speed")].selected == "fan-speed-2-100"


@pytest.mark.asyncio
async def test_set_state_in_f_force_c(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(portable_ac)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(
        portable_ac_id,
        hvac_mode="cool",
        target_temperature=24.5,
        selects={("fan-speed", "ac-fan-speed"): "fan-speed-2-100"},
        is_celsius=True,
    )
    await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller.items[0]
    assert dev.target_temperature_cooling.value == 24.5
    assert dev.hvac_mode.mode == "cool"
    assert dev.hvac_mode.previous_mode == "auto-cool"
    assert dev.selects[("fan-speed", "ac-fan-speed")].selected == "fan-speed-2-100"


@pytest.mark.asyncio
async def test_set_state_invalid_dev(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(portable_ac)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(
        "nope",
        hvac_mode="cool",
        target_temperature=22.5,
        selects={("fan-speed", "ac-fan-speed"): "fan-speed-2-100"},
    )
    mocked_controller._bridge.request.assert_not_called()


@pytest.mark.asyncio
async def test_set_state_no_updates(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(portable_ac)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(
        portable_ac_id,
    )
    mocked_controller._bridge.request.assert_not_called()


@pytest.mark.asyncio
async def test_set_state_invalid_updates(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(portable_ac)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(
        portable_ac_id,
        hvac_mode="i dont exist",
        selects={(None, None): 7, ("fan-speed", "ac-fan-speed"): "fan-speed-2-100"},
    )
    await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller.items[0]
    assert dev.selects[("fan-speed", "ac-fan-speed")].selected == "fan-speed-2-100"
